﻿using System;
using System.Drawing;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace AyaClock.NET
{
    /// <summary>
    /// 画面上の配置タイプ
    /// シリアライズを使ってセーブ・ロードするのでpublicにしておく
    /// </summary>
    public enum PositionType
    {
        LeftTop,
        LeftBottom,
        RightTop,
        RightBottom,
        Any
    }

    /// <summary>
    /// Windowの位置を調整する関数
    ///   AWindow  .......... 対象となるWindow
    ///   APosType .......... 配置する位置
    ///   AnIgnoreTaskBar ... Taskbarを無視するなら真
    /// 注意・XPでは常にTaskbarを無視する
    /// </summary>
    class FormUtils
    {
        [DllImport("user32.dll")]
        private extern static IntPtr FindWindow(string lpClassName, string lpWindowName);

        [StructLayout(LayoutKind.Sequential, Pack = 4)]
        private struct RECT
        {
            public int left;
            public int top;
            public int right;
            public int bottom;
        }

        [DllImport("User32.Dll")]
        private static extern bool GetWindowRect(IntPtr hWnd, ref RECT rect);

        static private bool IsWin8OrLater()
        {
            Version OSver = Environment.OSVersion.Version;
            return (OSver.Major > 6) ||
                   ((OSver.Major == 6) && (OSver.Minor >= 2));
        }

        static private bool GetTaskbarRect(ref RECT rect)
        {
            bool result = false;
            IntPtr hwnd = FindWindow("Shell_TrayWnd", null );
            if (hwnd != IntPtr.Zero)
            {
                result = GetWindowRect(hwnd, ref rect);
            }
            return result;
        }

        static public void AdjustWindowPos(Window AWindow, PositionType APosType, bool AnIgnoreTaskBar)
        {
            System.Drawing.Rectangle ownrc = new System.Drawing.Rectangle((Int32)AWindow.Left,
                                                                          (Int32)AWindow.Top,
                                                                          (Int32)AWindow.Width,
                                                                          (Int32)AWindow.Height);
            System.Windows.Forms.Screen scr = System.Windows.Forms.Screen.FromRectangle(ownrc);
            // defaultでは全画面が作業領域だと仮定
            RECT rc = new RECT();
            rc.left =  scr.Bounds.Left;
            rc.right = scr.Bounds.Right;
            rc.top = scr.Bounds.Top;
            rc.bottom = scr.Bounds.Bottom;
            // Windows8以降のタスクバーはSecondary monitorにも存在するので別処理
            if (IsWin8OrLater())
            {
                // for Win8以降
                if (!AnIgnoreTaskBar)
                {
                    Rectangle workrc = System.Windows.Forms.Screen.GetWorkingArea(ownrc);
                    rc.top = workrc.Top; rc.left = workrc.Left;
                    rc.bottom = workrc.Bottom; rc.right = workrc.Right;
                }
            } else {
                // for Win7まで
                if (!AnIgnoreTaskBar && scr.Primary)
                {
                    // Taskbarを考慮する場合
                    RECT workrc = new RECT();
                    if (GetTaskbarRect(ref workrc))
                    {
                        int w = (rc.left + rc.right) / 2, h = (rc.top + rc.bottom) / 2;
                        // Taskbarの位置を推定
                        if (workrc.right < w)
                        {
                            // Taskbarは左
                            if (workrc.right > 0) rc.left = workrc.right;
                        }
                        else if (workrc.left > w)
                        {
                            // Taskbarは右
                            if (workrc.left < rc.right) rc.right = workrc.left;
                        }
                        else if (workrc.bottom < h)
                        {
                            // Taskbarは上
                            if (workrc.bottom > 0) rc.top = workrc.bottom;
                        }
                        else
                        {
                            // Taskbarは下
                            if (workrc.top < rc.bottom) rc.bottom = workrc.top;
                        }
                    }
                    else
                    {
                        // 上手くTaskbar領域を取得できなかったのでWorkingAreaをTaskbar除外領域とする
                        rc.left = scr.WorkingArea.Left;
                        rc.right = scr.WorkingArea.Right;
                        rc.top = scr.WorkingArea.Top;
                        rc.bottom = scr.WorkingArea.Bottom;
                    }
                }
            }
            // 画面のDPIを参考にして調整
            rc.left = (int)DeviceHelper.DeviceToWPF(rc.left, Orientation.Horizontal);
            rc.right = (int)DeviceHelper.DeviceToWPF(rc.right, Orientation.Horizontal);
            rc.top = (int)DeviceHelper.DeviceToWPF(rc.top, Orientation.Vertical);
            rc.bottom = (int)DeviceHelper.DeviceToWPF(rc.bottom, Orientation.Vertical);
            //
            switch (APosType)
            {
                case PositionType.LeftTop:
                    AWindow.Left = rc.left;
                    AWindow.Top = rc.top;
                    break;
                case PositionType.LeftBottom:
                    AWindow.Left = rc.left; 
                    AWindow.Top = rc.bottom - AWindow.Height;
                    break;
                case PositionType.RightTop:
                    AWindow.Left = rc.right - AWindow.Width;
                    AWindow.Top = rc.top;
                    break;
                case PositionType.RightBottom:
                    AWindow.Left = rc.right - AWindow.Width;
                    AWindow.Top = rc.bottom - AWindow.Height;
                    break;
                case PositionType.Any:
                    if ((AWindow.Left+AWindow.Width) >= rc.right) AWindow.Left = rc.right - AWindow.Width;
                    if ((AWindow.Top + AWindow.Height) >= rc.bottom) AWindow.Top = rc.bottom - AWindow.Height;
                    break;
            }
            // ウインドウが画面をはみ出していた時のフォロー
            if ((AWindow.Left + AWindow.Width) >= rc.right)
            {
                AWindow.Left = rc.right - AWindow.Width;
            }
            if (AWindow.Left < rc.left)
            {
                AWindow.Left = rc.left;
            }
            if ((AWindow.Top + AWindow.Height) >= rc.bottom)
            {
                AWindow.Top = rc.bottom - AWindow.Height;
            }
            if (AWindow.Top < rc.top)
            {
                AWindow.Top = rc.top;
            }
        }
    }
}
