﻿using System;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace AyaClock.NET
{
    /// <summary>
    /// 時刻変更特権を持っているかどうかを確認するためのクラス
    ///   元ソース : http://gwave-systems.com/category.htm
    ///              http://bytes.com/topic/c-sharp/answers/225065-how-call-win32-native-api-gettokeninformation-using-c 
    ///              http://eternalwindows.jp/security/securitycontext/securitycontext04.html
    /// </summary>
    public class SystemTimePrivilege
    {
        /// <summary>
        /// 情報取得用トークン
        /// </summary>
        private const uint TOKEN_QUERY = 0x8;

        /// <summary>
        /// 特権を変更できるトークン
        /// </summary>
        private const uint TOKEN_ADJUST_PRIVILEGES = 0x20;

        /// <summary>
        /// 特権を有効にする値
        /// </summary>
        private const uint SE_PRIVILEGE_ENABLED = 0x2;

        /// <summary>
        /// 時刻変更用の特権名
        /// </summary>
        private const string SE_SYSTEMTIME_NAME = "SeSystemtimePrivilege";

        const int ERROR_NO_MORE_ITEMS = 259;

        enum TOKEN_INFORMATION_CLASS
        {
            TokenUser = 1,
            TokenGroups,
            TokenPrivileges,
            TokenOwner,
            TokenPrimaryGroup,
            TokenDefaultDacl,
            TokenSource,
            TokenType,
            TokenImpersonationLevel,
            TokenStatistics,
            TokenRestrictedSids,
            TokenSessionId,
            TokenGroupsAndPrivileges,
            TokenSessionReference,
            TokenSandBoxInert,
            TokenAuditPolicy,
            TokenOrigin,
            TokenElevationType,
            TokenLinkedToken,
            TokenElevation,
            TokenHasRestrictions,
            TokenAccessInformation,
            TokenVirtualizationAllowed,
            TokenVirtualizationEnabled,
            TokenIntegrityLevel,
            TokenUIAccess,
            TokenMandatoryPolicy,
            TokenLogonSid,
            vMaxTokenInfoClass
        }

        public enum TOKEN_ELEVATION_TYPE
        {
            TokenElevationTypeDefault = 1,
            TokenElevationTypeFull,
            TokenElevationTypeLimited
        }

        /// <summary>
        /// 現在のプロセス情報を取得します
        /// </summary>
        [DllImport("kernel32")]
        private static extern IntPtr GetCurrentProcess();

        /// <summary>
        /// プロセスのアクセストークンを開きます
        /// </summary>
        [DllImport("advapi32", SetLastError = true)]
        private static extern bool OpenProcessToken(IntPtr processHandle, uint desiredAccess, ref IntPtr tokenHandle);

        /// <summary>
        /// ハンドルを閉じます
        /// </summary>
        [DllImport("kernel32", SetLastError = true)]
        private static extern bool CloseHandle(IntPtr hHandle);

        /// <summary>
        /// 指定された特権名のLUID（ローカル一意識別子）を取得します
        /// </summary>
        [DllImport("advapi32", SetLastError = true, CharSet = CharSet.Unicode, EntryPoint = "LookupPrivilegeValueW")]
        private static extern bool LookupPrivilegeValue(string systemName, string name, ref long luid);

        /// <summary>
        /// LUID（ローカル一意識別子）と属性を格納する構造体
        /// </summary>
        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        private struct LUID_AND_ATTRIBUTES
        {
            public long Luid;
            public uint Attributes;
        }

        /// <summary>
        /// アクセストークン構造体
        /// </summary>
        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        private struct TOKEN_PRIVILEGES
        {
            public uint PrivilegeCount;
            public LUID_AND_ATTRIBUTES Privilege;
        }

        /// <summary>
        /// 指定したアクセストークン内の特権を有効または無効にします
        /// </summary>
        [DllImport("advapi32",SetLastError=true)]
        private static extern bool AdjustTokenPrivileges(IntPtr tokenHandle, bool disableAllPrivilege, ref TOKEN_PRIVILEGES newState, uint bufferLength, ref TOKEN_PRIVILEGES previousState, IntPtr returnLength);

        /// <summary>
        /// Tokenから各種情報を取得します
        /// </summary>
        [DllImport("advapi32", CharSet = CharSet.Auto)]
        static extern bool GetTokenInformation(IntPtr hToken,
                                               TOKEN_INFORMATION_CLASS tokenInfoClass,
                                               IntPtr TokenInformation,
                                               int tokeInfoLength,
                                               ref int reqLength);

        /// <summary>
        /// 現在のユーザーのタイプを返す。
        /// ノーマルユーザー・アドミン制限ユーザー・アドミンの３種類
        /// 取得に失敗した場合に偽を返す
        /// </summary>
        public static bool GetElevationType(ref TOKEN_ELEVATION_TYPE AnElevationType)
        {
            bool result = false;
            IntPtr hToken = IntPtr.Zero;
            if (OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY, ref hToken))
            {
                try
                {
                    int sz = 0;
                    IntPtr work = Marshal.AllocHGlobal(sizeof(TOKEN_ELEVATION_TYPE));
                    try
                    {
                        result = GetTokenInformation(hToken, TOKEN_INFORMATION_CLASS.TokenElevationType, work, sizeof(TOKEN_ELEVATION_TYPE), ref sz);
                        if (result)
                        {
                            AnElevationType = (TOKEN_ELEVATION_TYPE)Marshal.ReadInt32(work);
                        }
                    }
                    finally
                    {
                        Marshal.FreeHGlobal(work);
                    }
                }
                finally
                {
                    CloseHandle(hToken);
                }
            }
            return result;
        }

        /// <summary>
        /// 時刻変更特権を持っているかどうかを返す。有効かどうかではないので注意。
        /// 持っていれば真
        /// administratorならば常に真
        /// </summary>
        public static bool HaveSeSystemtimePrivilege()
        {
            // check admin
            TOKEN_ELEVATION_TYPE et = new TOKEN_ELEVATION_TYPE();
            if (GetElevationType(ref et))
            {
                if (et != TOKEN_ELEVATION_TYPE.TokenElevationTypeDefault) return true;
            }
            // check privilege
            bool sts = false;
            long luid = 0;
            if (LookupPrivilegeValue(null, SE_SYSTEMTIME_NAME, ref luid))
            {
                IntPtr hToken = IntPtr.Zero;
                if (OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY, ref hToken))
                {
                    try
                    {
                        int bufsz = 0;
                        GetTokenInformation(hToken,
                                            TOKEN_INFORMATION_CLASS.TokenPrivileges,
                                            IntPtr.Zero,
                                            0,
                                            ref bufsz);
                        if (bufsz > 0)
                        {
                            IntPtr tpwork = Marshal.AllocHGlobal(bufsz);
                            try
                            {
                                if (GetTokenInformation(hToken, TOKEN_INFORMATION_CLASS.TokenPrivileges, tpwork, bufsz, ref bufsz))
                                {
                                    TOKEN_PRIVILEGES tp = (TOKEN_PRIVILEGES)Marshal.PtrToStructure(tpwork, typeof(TOKEN_PRIVILEGES));
                                    for (int i = 0; i < tp.PrivilegeCount; i++)
                                    {
                                        /// <summary>
                                        /// using BLACK MAGIC.... for pointer calculation
                                        /// </sumamry>
                                        LUID_AND_ATTRIBUTES laa = (LUID_AND_ATTRIBUTES)Marshal.PtrToStructure((IntPtr)((int)tpwork + 4 + i * 12), typeof(LUID_AND_ATTRIBUTES));
                                        if (laa.Luid == luid)
                                        {
                                            sts = true; // SeSystemtimePrivilege発見
                                            break;
                                        }
                                    }
                                }
                            }
                            finally
                            {
                                Marshal.FreeHGlobal(tpwork);
                            }
                        }
                    }
                    finally
                    {
                        CloseHandle(hToken);
                    }
                }
            }
            return sts;
        }
    }
}